// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2025 Oracle.  All Rights Reserved.
 * Author: Darrick J. Wong <djwong@kernel.org>
 */
use std::io;
use clap::Parser;
use std::fs::File;
use xfs_healer::healthmon;
use xfs_healer::softhandle;
use xfs_healer::repair;
use xfs_healer::fsgeom;
use xfs_healer::fsprops;

/// Interpret command line arguments
#[derive(Parser, Debug)]
struct Cli {
    /// Check to see if monitoring is supported.
    #[arg(short, long)]
    check: bool,

    /// Enable debugging messages.
    #[arg(short, long)]
    debug: bool,

    /// Log health events to stdout.
    #[arg(short, long)]
    log: bool,

    /// Capture all events.
    #[arg(short, long)]
    everything: bool,

    /// Repair broken metadata?
    #[arg(short, long)]
    repair: bool,

    /// Decide if we'll repair broken metadata using fs property?
    #[arg(short, long)]
    autofsck: bool,

    /// XFS filesystem mountpoint to monitor.
    path: std::path::PathBuf,
}

fn __main(args: &mut Cli) -> io::Result<i32> {
    if args.debug {
        println!("{:?}", args);
    }

    let fp = File::open(&args.path)?;

    /* Just a presence check? */
    if args.check {
        match healthmon::is_supported(&fp) {
            Ok(_)  => std::process::exit(0),
            Err(_) => std::process::exit(1),
        }
    }

    /* Complain a bit if repairs won't be entirely effective. */
    let fsgeom = fsgeom::geometry(&fp)?;
    if args.repair {
        if ! fsgeom::xfs_has_rmapbt(&fsgeom) {
            println!("{}: XFS online repair is less effective without rmap btrees",
                     args.path.display());
        }
        if ! fsgeom::xfs_has_parent(&fsgeom) {
            println!("{}: XFS online repair is less effective without parent pointers",
                     args.path.display());
        }
    }

    let fh = softhandle::create(&fp, &args.path)?;

    /*
     * Caller wants us to figure out if we're doing repairs from the admin
     * configurable fs property
     */
    if args.autofsck {
        match fsprops::getstr(&fp, fsprops::AUTOFSCK_NAME) {
            Some(value) => {
                if value == "repair" {
                    args.repair = true;
                }
            },
            _ => { }
        }
    }
    let hmon = healthmon::create(fp, &args.path, args.everything, args.debug)?;

    for f in hmon {
        if args.log {
            f.log();
        }
        if args.repair {
            match f.schedule_repair() {
                Some(vec) => {
                    for repair in vec {
                        repair::repair(&fh, repair)
                    }
                },
                _ => continue,
            }
        }
    }

    Ok(0)
}

fn main() {
    let mut args = Cli::parse();

    match __main(&mut args) {
        Ok(f)   => std::process::exit(f),
        Err(e)  => {
            eprintln!("{}: {}", args.path.display(), e);
            std::process::exit(1);
        }
    };
}
