// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2025 Oracle.  All Rights Reserved.
 * Author: Darrick J. Wong <djwong@kernel.org>
 */
use std::io;
use crate::xfs_fs;
use crate::softhandle;
use std::os::fd::AsRawFd;
use nix::ioctl_readwrite;

ioctl_readwrite!(xfs_ioc_scrub_metadata, 'X', 60, xfs_fs::xfs_scrub_metadata);

/// Schedule a full-filesystem repair
pub fn create_fs_repair(sm_type: u32) ->
        xfs_fs::xfs_scrub_metadata {
    xfs_fs::xfs_scrub_metadata {
        sm_type,
        sm_flags: xfs_fs::XFS_SCRUB_IFLAG_REPAIR,
        sm_ino: 0,
        sm_gen: 0,
        sm_agno: 0,
        sm_reserved: [0, 0, 0, 0, 0],
    }
}

/// Schedule a file repair
pub fn create_file_repair(sm_type: u32, ino: u64, gen: u32) ->
        xfs_fs::xfs_scrub_metadata {
    xfs_fs::xfs_scrub_metadata {
        sm_type,
        sm_flags: xfs_fs::XFS_SCRUB_IFLAG_REPAIR,
        sm_ino: ino,
        sm_gen: gen,
        sm_agno: 0,
        sm_reserved: [0, 0, 0, 0, 0],
    }
}

/// Schedule a group repair
pub fn create_group_repair(sm_type: u32, group: u32) ->
        xfs_fs::xfs_scrub_metadata {
    xfs_fs::xfs_scrub_metadata {
        sm_type,
        sm_flags: xfs_fs::XFS_SCRUB_IFLAG_REPAIR,
        sm_ino: 0,
        sm_gen: 0,
        sm_agno: group,
        sm_reserved: [0, 0, 0, 0, 0],
    }
}

fn __repair(fh: &softhandle::SoftHandle, mut blob: xfs_fs::xfs_scrub_metadata)
        -> io::Result<bool> {
    let fp = fh.reopen()?;

    unsafe { xfs_ioc_scrub_metadata(fp.as_raw_fd(), &mut blob)?; }
    Ok(true)
}

/// Try to repair something, or log whatever went wrong
pub fn repair(fh: &softhandle::SoftHandle, blob: xfs_fs::xfs_scrub_metadata) {
    match __repair(fh, blob) {
        Err(e)  => {
            eprintln!("{}: {}", fh, e);
        },
        _ => return,
    };
}
