// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2025 Oracle.  All Rights Reserved.
 * Author: Darrick J. Wong <djwong@kernel.org>
 */
use std::io;
use std::ptr;
use std::fs::File;
use crate::libhandle;
use libhandle::size_t;
use std::os::fd::AsRawFd;
use std::os::raw::{c_void, c_int};

// dumb wrappers around libhandle functions and data structures

fn fd_to_handle(
        fd: c_int,
        hanp: *mut *mut c_void,
        hlen: *mut size_t
    ) -> nix::Result<c_int> {
    unsafe {
        nix::errno::Errno::result(libhandle::fd_to_handle(fd, hanp, hlen))
    }
}

fn free_handle(hanp: *mut c_void, hlen: size_t) {
    unsafe {
        libhandle::free_handle(hanp, hlen)
    }
}

struct Handle {
    /// Opaque handle data
    hanp: *mut c_void,

    /// Size of handle
    hlen: size_t,
}

impl Drop for Handle {
    fn drop(&mut self) {
        free_handle(self.hanp, self.hlen)
    }
}

extern {
    fn memcmp(s1: *const c_void, s2: *const c_void, n: size_t) -> c_int;
}

impl PartialEq for Handle {
    fn eq(&self, other: &Self) -> bool {
        if self.hlen != other.hlen {
            return false
        }

        unsafe {
            memcmp(self.hanp, other.hanp, self.hlen) == 0
        }
    }
}

/// Create a Handle for an open file
impl Handle {
    pub fn from_file(fp: &File) -> io::Result<Handle> {
        let mut hanp = ptr::null_mut();
        let mut hlen: size_t = 0;

        fd_to_handle(fp.as_raw_fd(), &mut hanp, &mut hlen)?;
        Ok(Handle { hanp, hlen })
    }
}

/// Soft handle that can be reconnected to a filesystem
pub struct SoftHandle {
    /// path to the filesystem mountpoint
    mountpoint: std::path::PathBuf,

    /// Filesystem handle
    handle: Handle,
}

impl SoftHandle {
    /// Try to reopen the filesystem from which we got the handle.
    pub fn reopen(&self) -> io::Result<File> {
        let fp = File::open(&self.mountpoint)?;

        if Handle::from_file(&fp)? != self.handle {
            let s = format!("{}: Stale file handle", self.mountpoint.display());
            return Err(io::Error::new(io::ErrorKind::Other, s))
        }

        Ok(fp)
    }
}

impl std::fmt::Display for SoftHandle {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(f, "{}", self.mountpoint.display())
    }
}

/// Create a soft handle from an open file descriptor and its mount point
pub fn create(fp: &File, mountpoint: &std::path::PathBuf) ->
        io::Result<SoftHandle> {
    Ok(SoftHandle {
        mountpoint: mountpoint.clone(),
        handle: Handle::from_file(&fp)?,
    })
}
